(* Hasznlja a 3. kishziknt megrt helyese fggvnyt.
*)

app load ["Int", "StringCvt", "Time", "Timer"];
use "helyese.sml";
use "sudoku.sml";

type size  = int;
type field = int;
type board = field list list;
type sspec  = size * board;

exception Sudoku of string;

(* helyes_feladvany kt = igaz, ha a sudoku fggvny kt paramtere
                         kielgti ezt a prediktumot (a mezrtkek
                         helyessgt beolvass kzben ellenrizzk)
   helyes_feladvany : sspec -> bool
*)
fun helyes_feladvany (kt as (k,t)) =
      let
	  val (mx, my) = (length(hd t), length t)
	  val (cx, cy) = (mx div k, my div k) 
      in
	  (* a cellk szma megfelel *)
	  1 <= k andalso k <= 5 andalso
	  (* minden sorban ugyanannyi mez van *)
	  List.all (fn x => x = mx) (map length t) andalso
 	  (* a mezk s cellk szma fgglegesen megfelel *)
	  my mod k = 0 andalso 1 <= cy andalso cy <= k andalso
 	  (* a mezk s cellk szma vzszintesen megfelel *)
	  mx mod k = 0 andalso 1 <= cx andalso cx <= k andalso
	  (* a tbla megfelel a sudoku-szablyoknak *)
	  helyese kt
      end (* let *);

(* sudoku_be f = az f szvegfjlbl beolvasott tbla
   sudoku_be : string -> sspec
*)
fun sudoku_be f =
      let
	  (* str2int c = a c fzr elejn ll karaktersorozat
                         egsz szmknt, ha konvertlhat,
                         egybknt Sudoku kivtelt jelez
	     str2int : string -> field
	  *)
	  fun str2int c =
	        case Int.fromString c of
		    NONE   => raise Sudoku("Nemvrt karaktersorozat: '" ^ c ^ "'.")
		  | SOME v => v
	  val istr = TextIO.openIn f
	  val k    = str2int(TextIO.inputLine istr)
	  (* str2fld c = a c fzr elejn ll karaktersorozat
                         egsz szmknt, ha konvertlhat s a
                         megengedett 0..k*k tartomnyba esik,
                         egybknt Sudoku kivtelt jelez
	     str2fld : string -> field
	  *)
	  fun str2fld c =
	       let
		   val v = str2int c
	       in
		   if 0 <= v andalso v <= k*k then v
		   else raise Sudoku("Hibs mezrtk: '" ^ c ^ "'.")
	       end (* let *)
	  val m  = TextIO.inputAll istr before TextIO.closeIn istr
	  val s  = String.tokens (fn c => c = #"\n") m
	  val t  = map (map str2fld o String.tokens Char.isSpace) s
	  val kt = (k, t) before TextIO.closeIn istr
      in
	  if helyes_feladvany kt then kt
	  else raise Sudoku("Hibs a feladvny a(z) " ^ f ^ " fjlban.")
      end (* let *);

(* brd2str kt = a kt tbla fzr alakban
   brd2str : sspec -> string
*)
fun brd2str (kt as (k,t)) =
      let
	  val (maxx, maxy) = (length(hd t), length t)
	  val h = size(Int.toString(k*k*k*k))
          (* row2str (i, xs) = egy sor mezrtkei fzrknt
	     row2str : size * field -> string
          *)
	  fun row2str (_, []) = ""
	    | row2str (i, (x::xs)) =
                (StringCvt.padLeft #" " h (Int.toString x) ^
		 (if i < maxx andalso i mod k = 0
		  then " "
		  else "") ^
		 row2str(i+1, xs)
		)
          (* col2str (i, xss) = az oszlopok tartalma sorfolytonos fzrknt
	     col2str : size * field list -> string
          *)
	  fun col2str (_, []) = "\n"
	    | col2str (i, (xs::xss)) =
                (row2str(1, xs) ^
		 (if i < maxy andalso i mod k = 0
		  then "\n\n"
		  else if i < maxy
		  then "\n"
		  else "") ^
		 col2str(i+1, xss)
		)
      in
	  col2str(1, t)
      end (* let *);

(* brds2str ms = az ms megoldslista elemei fzr alakban,
                 res sorokkal elvlasztva
   brds2str sspec list -> string
*)
fun brds2str ms =
      let
	  val sep = "\n-----\n\n"
      in
	  foldl (fn (m,s) => s ^ brd2str m ^ sep) sep ms
      end (* let *);

(* sudoku_ki(f, ms) = az ms megoldslista elemeit fzr alakban
                      kirja az f szveges llomnyba
   sudoku_ki : string * sspec list -> unit
*)
fun sudoku_ki(f, ms) =
      let val ostr = TextIO.openOut f
	  val _ = TextIO.output(ostr, brds2str ms)
      in
	 TextIO.closeOut ostr
      end (* let *);

(* megold(f, m) = Beolvas egy feladvnyt az f szveges llomnybl s
      kirja az sszes megoldst az m szveges llomnyba. Ehhez
      szksge van a sudoku fggvnyre.  Eredmnye az f nevet s a
      megoldsok szmt tartalmaz fzr.
   megold : string * string -> string
 *)
fun megold (f, m) =
    let val kt = sudoku_be f
	val kts = sudoku kt
	val _ = sudoku_ki(m, kts)
    in
	"Feladvny: " ^ f ^ ", " ^ Int.toString(length kts) ^ " megolds\n"
    end (* let *);

(* stopper(f, m) = Beolvas egy feladvnyt az f szveges llomnybl s
      kirja az sszes megoldst az m szveges llomnyba. Ehhez
      szksge van a sudoku fggvnyre.  Eredmnye az f nevet, a
      megoldsok szmt s a futsi idt mp-ben tartalmaz fzr.
   stopper : string * string -> string
 *)
fun stopper (f, m) =
    let val kt = sudoku_be f
	val start = Timer.startCPUTimer()
	val kts = sudoku kt
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = sudoku_ki(m, kts)
    in
	"Feladvny: " ^ f ^ ", " ^
	Int.toString(length kts) ^ " megolds, " ^
	"id: " ^ Time.fmt 2 tim ^ " mp\n"
    end (* let *);
